/*
 *  rareVariantsAnalysis.cpp
 *  LogitReg
 *
 *  Created by Youna Hu on 1/3/11.
 *  Copyright 2011 U of Michigan. All rights reserved.
 *
 */

#include "rareVariantsAnalysis.h"

//gsl_rng * RareVariantAnalysis::r;

void RareVariantAnalysis::Initializer(Matrix & xGene, Vector & yIn, int nPermutation)
{
  p = 0.0;
  statistics = -1.0;
	twoBytwoMat.InitializeFromNums(0,0,0,0);
	yPhenoReg = yIn;
	yControl1 = yPhenoReg;
	// remember that yControl1 denotes controls, so yControl1 = 1- yPhenoReg;
	yControl1.Negate();
	yControl1 += 1.0;
	// genotype matrix -- deal with the case that the genetic score is 1 for every individual
	nVariants = xGene.cols;
	if (nVariants == 0) {
		printf("Error: Emptye xGene matrix!\n");
		throw;
	}
	xGeno = xGene;
	// sample size
	nSample = yIn.Length();
	xIntercept.Dimension(nSample, 1, 1); // a nSample by 1 matrix with elements as 1
	xGenoCov = xIntercept;
	rvWeight.Dimension(nVariants);
	xGenoVariant.Dimension(nSample);
	xGenoCombinedMat.Dimension(nSample,1);
	/*for (int i = 0; i < nSample; i ++) {
		printf("xGenoCombineMat[i][0] = %1.3f\n",xGenoCombinedMat[i][0]);
	}*/
	sameScore = true;
	// Permutation parameters
	nPermute = nPermutation;
	orderPermute.resize(nSample);
	for (int i = 0; i < nSample; i ++) {
		orderPermute.push_back(i);
	}
	yPhenoPermute.Dimension(nSample);
	statPermute.Dimension(nPermute);
	// need to record the 5% and 95% p values for the QQplot since the p does not follow a uniform distribution any more
  // pRangePermute.Dimension(nPermute,2); 
}

void RareVariantAnalysis::Initializer(Matrix &xGene, Vector &yIn, std::string &choiceCollapse, int nPermutation){
	Initializer(xGene,yIn,nPermutation);
	if (choiceCollapse != "or" && choiceCollapse != "sum" && choiceCollapse != "wt") {
		error("Invalid collapsing choice. Pick one from {or, sum, wt}\n");
	}	
}

void RareVariantAnalysis::Initializer(Matrix & xGene, Matrix &xCovIn, Vector & yIn, int nPermutation)
{
	Initializer(xGene, yIn, nPermutation);
 	xCov = xCovIn;
}


void RareVariantAnalysis::Initializer(Matrix & xGene, Matrix &xCovIn, Vector & yIn, std::string &choiceCollapse, int nPermutation)
{
	Initializer(xGene, xCovIn, yIn, nPermutation);
	if (choiceCollapse != "or" || choiceCollapse != "sum" || choiceCollapse != "wt") {
		error("Invalid collapsing choice. Pick one from {or, sum, wt}\n");
	}
}



RareVariantAnalysis::~RareVariantAnalysis(){
}

RareVariantAnalysis::RareVariantAnalysis(){
}

RareVariantAnalysis::RareVariantAnalysis(Matrix & xGene, Vector & yIn, int nPermutation){
	Initializer(xGene, yIn, nPermutation); // if nPermutation isn't specified, the use default
}

RareVariantAnalysis::RareVariantAnalysis(Matrix & xGene, Vector & yIn, std::string &choiceCollapse, int nPermutation){
	Initializer(xGene, yIn, choiceCollapse, nPermutation); // if nPermutation isn't specified, the use default
}

RareVariantAnalysis::RareVariantAnalysis(Matrix & xGene, Matrix &xCovIn, Vector & yIn, int nPermutation){
	Initializer(xGene, xCovIn, yIn, nPermutation); // if nPermutation isn't specified, the use default
}

RareVariantAnalysis::RareVariantAnalysis(Matrix & xGene, Matrix &xCovIn, Vector & yIn, std::string & choiceCollapse, int nPermutation){
	Initializer(xGene, xCovIn, yIn, choiceCollapse, nPermutation); // if nPermutation isn't specified, the use default
}


void RareVariantAnalysis::CalcFreqCtrl(bool smallSampleAdjust) // can easily manipulate yControl1 to calculate frequency in case, or overall
{ // Calculate Freqency in Ctrl with small sample adjustment
	// note: We use xGenoCombined to store each column of the genotype matrix and calculate the 
	for (int i = 0; i < nVariants; i ++) {
		// store each column
		for (int j = 0; j < nSample; j ++) {
			xGenoVariant[j] = xGeno[j][i];
		}
		// Then calcualte the dot product with (1-yPhenoReg) so as to calculate the allele frequency in control
		// note: Need to remember that yPhenoReg has been changed here. Need to invert it back at the regression step
		if (smallSampleAdjust) {
			rvWeight[i] = (xGenoVariant.InnerProduct(yControl1) + 1.0)/(2.0*nSample + 2.0); // with same sample correction		
		} else {
			rvWeight[i] = (xGenoVariant.InnerProduct(yControl1))/(2.0*nSample); // without same sample correction		
		}

	}		
}

void RareVariantAnalysis::WeightMadsonBrowning(){
	// we need to use the allele frequency in control	
	CalcFreqCtrl(true);
}

void RareVariantAnalysis::SetWeight(Vector & userWeight){
	rvWeight = userWeight;
}

// Update the xGenoCombined vector of rare variant scores
void RareVariantAnalysis::CollapseType(std::string choice, Vector &userWeight){ // update xGenoCombined
	
	if (choice == "or") { // if at any of the position, the subjects has an alternative copy, the genetic score is 1, otherwise, 0
		// if the sum of each row is > zero, then the score is 1, otherwise, 0		
		for (int i = 0; i < nSample; i ++) {
			xGenoCombinedMat[i][0] = (xGeno[i].Sum() > 0.0);
			sameScore = sameScore && (xGenoCombinedMat[i][0] == xGenoCombinedMat[0][0]);
		//	printf("sum = %f, xGenoCombinedMat[%d][0] = %f, sameScore = %d \n",xGeno[i].Sum(), i, xGenoCombinedMat[i][0],sameScore);
		}
	} else if (choice == "sum") {
		for (int i = 0; i < nSample; i ++) {
			xGenoCombinedMat[i][0] = xGeno[i].Sum();
			sameScore = sameScore && (xGenoCombinedMat[i][0] == xGenoCombinedMat[0][0]);
		}
	} else if (choice == "wt"){
		//update the weights
		WeightMadsonBrowning();
		for (int i = 0; i < nSample; i ++) {
			xGenoCombinedMat[i][0] = xGeno[i].InnerProduct(rvWeight);
			sameScore = sameScore && (xGenoCombinedMat[i][0] == xGenoCombinedMat[0][0]);
		}
	} else if (choice == "user") {
		// update the user specified weights
		SetWeight(userWeight);
		for (int i = 0; i < nSample; i ++) {
			xGenoCombinedMat[i][0] = xGeno[i].InnerProduct(rvWeight);
			sameScore = sameScore && (xGenoCombinedMat[i][0] == xGenoCombinedMat[0][0]);
		}
	} else {
		error("Invalid collapsing choice. It should be one of {or, sum, wt}\n");
	}

	if (sameScore) {
		printf("Warning: The rare variant score is the same for every subject!\n");
	}
}

// this version of RVregression only considers rare variants score as covariates
void RareVariantAnalysis::RVregression(std::string choice, Vector & userWeight){ // need intercept, xGenoCombined, covariates
	// first update xGenoCombined using CMC
	CollapseType("or",userWeight);
	// if the genetic score is the same for every individual, the 
	if (sameScore) {
		p = 1.0;
		return;
	} 
	// check to see if Fisher's exact test should be used 
	twoBytwoMat.InitializeFromVecMat(yPhenoReg, xGenoCombinedMat); // first initialize the 2 by 2 table
	// twoBytwoMat.Print();
	
	if (twoBytwoMat.MustExactTest()){
		twoBytwoMat.FisherExactTest();
		p = twoBytwoMat.GetpExact();
	} else {
		// update the right xGenoCombined
		if (choice != "or") {
			CollapseType(choice,userWeight);
		}
		LogisticRegression lr;
		int nrrounds = 20;
		// need to make a matrix with columns(ones,xGenoCombined)
		xGenoCov.StackLeft(xGenoCombinedMat);
	// 	printf("xGenoCov size (%d,%d)\n", xGeno.rows,xGeno.cols);
		if (lr.FitLogisticModel(xGenoCov, yPhenoReg, nrrounds)) {
			p = lr.GetAsyPvalue()[1];
		} else {
			printf("Error in logistic regression!\n");
		}
	}
}

void RareVariantAnalysis::RVregressionWithCov(std::string choice, Vector &userWeight){
	CollapseType("or",userWeight);
	if (sameScore) {
		p = 1.0;
	}
	twoBytwoMat.InitializeFromVecMat(yPhenoReg, xGenoCombinedMat);
	
	if (choice != "or"){
		CollapseType(choice,userWeight);
	}
	// check to see if exact logistic regression should be used ... ... 
	// for now, simplye use the regression routine
	
	xGenoCov.StackLeft(xGenoCombinedMat);
	xGenoCov.StackLeft(xCov);
	LogisticRegression lr;
	int nrrounds = 20;
	lr.FitLogisticModel(xGenoCov, yPhenoReg, nrrounds);
	p = lr.GetAsyPvalue()[1]; // or, we can use likelihood test here, refer: RVregPermuteObsWithCov()
}

// Is this test a two sided or one sided? (default = Yes)
// Is the smaller statistics, the better it is? (default = Yes)
void RareVariantAnalysis::Pvalue(bool twoSided, bool smallerIsExtremer)
{
	// smallerIsBetter means that the extreme is small values
	if (smallerIsExtremer) {
		// p value -- 
		p = ((double) (nPermute - statPermute.CountIfGreater(statistics)))/nPermute;
		if (twoSided) {
			if (p <= 0.5) {
				p = 2*p;
			} else {
				p = 2*(1-p);
			}
		}
	} else { // statitistics -- biggerIsBetter
		p = ((double) statPermute.CountIfGreater(statistics))/nPermute;
		if (twoSided) {
			if (p <= 0.5) {
				p = 2*p;
			} else {
				p = 2*(1-p);
			}
		}
	} 
}

// find the observed statistics
void RareVariantAnalysis::RVregPermuteObs(std::string choice, Vector &userWeight){
	// first update xGenoCombined (using CMC)
	CollapseType("or",userWeight);
	twoBytwoMat.InitializeFromVecMat(yPhenoReg, xGenoCombinedMat);
	if (choice != "or") {
		CollapseType(choice, userWeight);
	}
	if (sameScore) {
		p = 1.0;
		return;
	}
	
	LogisticRegression lr;
	int nrrounds = 20;
	xGenoCov.StackLeft(xGenoCombinedMat); 
	if (lr.FitLogisticModel(xGenoCov, yPhenoReg, nrrounds)) {
		p = lr.GetAsyPvalue()[1]; // Note, here, I used the wald test to calculate the p value
															// A more stricter way is to be consistent and use the likelihood ratio test.
															// The two ways should be equivalent and won't too much difference. 
		statistics = lr.GetDeviance(xGenoCov, yPhenoReg);
		lr.FitLogisticModel(xIntercept, yPhenoReg, nrrounds); //the smaller the number of covariates is, the bigger the deviance is
		statistics = lr.GetDeviance(xIntercept, yPhenoReg) - statistics;		
	} else {
		printf("Error in logistic regression!\n");
	}
}

void RareVariantAnalysis::RVregPermuteObsWithCov(std::string choice, Vector &userWeight){
	// first update xGeno
	CollapseType(choice,userWeight);
	if (sameScore) {
		p = 1.0;
		return;
	}
	LogisticRegression lr;
	int nrrounds = 20;
	xGenoCov.StackLeft(xGenoCombinedMat);
	xGenoCov.StackLeft(xCov);
	if (lr.FitLogisticModel(xGenoCov, yPhenoReg, nrrounds)) {
		p = lr.GetAsyPvalue()[1];
		statistics = lr.GetDeviance(xGenoCov, yPhenoReg);
//		printf("statistics = %f\n", statistics);
		lr.FitLogisticModel(xInterceptCov, yPhenoReg, nrrounds);
		statistics = lr.GetDeviance(xInterceptCov, yPhenoReg) - statistics;
	//	printf("statistics = %f\n", statistics);
	} else{
		printf("Error in logistic regression!\n");
	}	
}

// using the same covariate matrix, regress it again a permuated outcome
void RareVariantAnalysis::RVregPermuteEach(std::string choice, Vector &userWeight){
	// Assume that both xGenoCov and xIntercept has been updated, 
	// the former in the RVregPermuteObs() function and the latter in the constructor
	if (sameScore) {
		p = 1.0;
		return;
	} 	
	LogisticRegression lr;
	int nrrounds = 20;
	if (lr.FitLogisticModel(xGenoCov, yPhenoReg, nrrounds)) {
		statistics = lr.GetDeviance(xGenoCov, yPhenoReg);
		lr.FitLogisticModel(xIntercept, yPhenoReg, nrrounds);
		statistics = lr.GetDeviance(xIntercept, yPhenoReg) - statistics;		
	} else {
		printf("Error in logistic regression!\n");
	}
}

void RareVariantAnalysis::RVregPermute(std::string choice, Vector &userWeight){
	if (nPermute == 0) {
		printf("Error: Permutation time = %d\n",nPermute);
		return;
	}
	if (sameScore) {
		printf("Warning: Uniform rare variant scores. No need to do permutation!\n");
		p = 1.0;
		return;
	}	
	// permutation loop
	for (int i = 0; i < nPermute; i++) {
		gsl_ran_shuffle(g_r, &(orderPermute[0]), nSample, sizeof(int));
		// update the phenotype at this permutation order
		for (int j = 0; j < nSample; j ++) {
			yPhenoPermute[j] = yPhenoReg[orderPermute[j]];
		}
		// use the permuted y for regression
		RVregPermuteEach(choice,userWeight);
		statPermute[i] = statistics;
	}
	Pvalue();
}

Matrix &	RareVariantAnalysis::GetGenoCombined(){
	for (int i = 0; i < nSample; i ++) {
		printf("xGenoCombinedMat[%d] = %f\n",i,xGenoCombinedMat[i][0]);
	}
	return (xGenoCombinedMat);
}

Vector & RareVariantAnalysis::GetWeights(){
	for (int i = 0; i < nVariants; i ++) {
		printf("rvWeight[%d] = %f\n",i,rvWeight[i]);
	}
	return (rvWeight);
}

double & RareVariantAnalysis::GetP(){
	printf("p = %f\n", p);
	return(p);
}

double & RareVariantAnalysis::GetStatistics(){
	printf("statitics = %f\n", statistics);
	return(p);
}



